/* Copyright (C) 2011, 2012, 2013 by Bundesanstalt für Gewässerkunde
 * Software engineering by Intevation GmbH
 *
 * This file is Free Software under the GNU AGPL (>=v3)
 * and comes with ABSOLUTELY NO WARRANTY! Check out the
 * documentation coming with Dive4Elements River for details.
 */

package org.dive4elements.river.exports;

import java.io.OutputStream;
import java.text.DateFormat;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import org.apache.log4j.Logger;
import org.dive4elements.artifacts.CallMeta;
import org.dive4elements.artifacts.common.utils.Config;
import org.dive4elements.river.artifacts.D4EArtifact;
import org.dive4elements.river.artifacts.access.RangeAccess;
import org.dive4elements.river.artifacts.access.RiverAccess;
import org.dive4elements.river.artifacts.model.CalculationResult;
import org.dive4elements.river.artifacts.model.WKmsJRDataSource;
import org.dive4elements.river.artifacts.model.WQDay;
import org.dive4elements.river.artifacts.resources.Resources;
import org.dive4elements.river.utils.Formatter;
import org.dive4elements.river.utils.RiverUtils;

import au.com.bytecode.opencsv.CSVWriter;
import net.sf.jasperreports.engine.JRException;
import net.sf.jasperreports.engine.JasperExportManager;
import net.sf.jasperreports.engine.JasperFillManager;
import net.sf.jasperreports.engine.JasperPrint;

/**
 * @author <a href="mailto:ingo.weinzierl@intevation.de">Ingo Weinzierl</a>
 */
public class DurationCurveExporter extends AbstractExporter {

    /** The log used in this exporter. */
    private static Logger log = Logger.getLogger(DurationCurveExporter.class);

    public static final String CSV_DURATION_HEADER = "export.duration.curve.csv.header.duration";

    public static final String CSV_W_HEADER = "export.duration.curve.csv.header.w";

    public static final String CSV_Q_HEADER = "common.export.csv.header.q";

    public static final String PDF_HEADER_MODE = "export.duration.pdf.mode";
    public static final String JASPER_FILE = "export.duration.pdf.file";

    /** The storage that contains all WQKms objects for the different facets. */
    protected List<WQDay> data;

    public DurationCurveExporter() {
        this.data = new ArrayList<>();
    }

    @Override
    protected void addData(Object d) {
        if (d instanceof CalculationResult) {
            d = ((CalculationResult) d).getData();
            if (d instanceof WQDay) {
                this.data.add((WQDay) d);
            }
        }
    }

    @Override
    protected void writeCSVData(final CSVWriter writer) {
        log.info("DurationCurveExporter.writeData");

        writeCSVHeader(writer);

        for (final WQDay wqday : this.data) {
            wQDay2CSV(writer, wqday);
        }
    }

    protected void writeCSVHeader(final CSVWriter writer) {
        log.info("DurationCurveExporter.writeCSVHeader");

        final String unit = new RiverAccess((D4EArtifact) this.master).getRiver().getWstUnit().getName();

        writer.writeNext(new String[] { msg(CSV_W_HEADER, new Object[] { unit }), msg(CSV_Q_HEADER), msg(CSV_DURATION_HEADER) });
    }

    protected void wQDay2CSV(final CSVWriter writer, final WQDay wqday) {
        log.debug("DurationCurveExporter.wQDay2CSV");

        final int size = wqday.size();

        final NumberFormat wf = getWFormatter();
        final NumberFormat qf = getQFormatter();
        final NumberFormat df = getDFormatter();

        if (wqday.isIncreasing()) {
            for (int i = size - 1; i >= 0; i--) {
                writer.writeNext(new String[] { wf.format(wqday.getW(i)), qf.format(wqday.getQ(i)), df.format(wqday.getDay(i)) });
            }
        } else {
            for (int i = 0; i < size; i++) {
                writer.writeNext(new String[] { wf.format(wqday.getW(i)), qf.format(wqday.getQ(i)), df.format(wqday.getDay(i)) });
            }
        }
    }

    /**
     * Returns the number formatter for W values.
     *
     * @return the number formatter for W values.
     */
    @Override
    protected NumberFormat getWFormatter() {
        return Formatter.getDurationW(this.context);
    }

    /**
     * Returns the number formatter for Q values.
     *
     * @return the number formatter for Q values.
     */
    @Override
    protected NumberFormat getQFormatter() {
        return Formatter.getDurationQ(this.context);
    }

    /**
     * Returns the number formatter for duration values.
     *
     * @return the number formatter for duration values.
     */
    protected NumberFormat getDFormatter() {
        return Formatter.getDurationD(this.context);
    }

    @Override
    protected void writePDF(final OutputStream out) {
        final WKmsJRDataSource source = createJRData();

        final String jasperFile = Resources.getMsg(this.context.getMeta(), JASPER_FILE, "/jasper/duration_en.jasper");
        final String confPath = Config.getConfigDirectory().toString();

        final Map parameters = new HashMap();
        parameters.put("ReportTitle", "Exported Data");
        try {
            final JasperPrint print = JasperFillManager.fillReport(confPath + jasperFile, parameters, source);
            JasperExportManager.exportReportToPdfStream(print, out);
        }
        catch (final JRException je) {
            log.warn("Error generating PDF Report!");
            je.printStackTrace();
        }
    }

    protected WKmsJRDataSource createJRData() {
        final WKmsJRDataSource source = new WKmsJRDataSource();

        addMetaData(source);
        for (final WQDay wqday : this.data) {
            addWQDayData(source, wqday);
        }

        return source;
    }

    protected void addMetaData(final WKmsJRDataSource source) {
        final CallMeta meta = this.context.getMeta();

        final D4EArtifact flys = (D4EArtifact) this.master;

        source.addMetaData("river", RiverUtils.getRivername(flys));

        final Locale locale = Resources.getLocale(meta);
        final DateFormat df = DateFormat.getDateInstance(DateFormat.SHORT, locale);

        source.addMetaData("date", df.format(new Date()));

        final RangeAccess rangeAccess = new RangeAccess(flys);
        final double[] kms = rangeAccess.getKmRange();
        source.addMetaData("range", String.valueOf(kms[0]));

        source.addMetaData("calculation", Resources.getMsg(locale, PDF_HEADER_MODE, "Duration"));
    }

    protected void addWQDayData(final WKmsJRDataSource source, final WQDay wqday) {
        final int size = wqday.size();

        final NumberFormat wf = getWFormatter();
        final NumberFormat qf = getQFormatter();
        final NumberFormat df = getDFormatter();

        if (wqday.isIncreasing()) {
            for (int i = size - 1; i >= 0; i--) {
                source.addData(new String[] { "", wf.format(wqday.getW(i)), qf.format(wqday.getQ(i)), "", "", "", df.format(wqday.getDay(i)) });
            }
        } else {
            for (int i = 0; i < size; i++) {
                source.addData(new String[] { "", wf.format(wqday.getW(i)), qf.format(wqday.getQ(i)), "", "", "", df.format(wqday.getDay(i)) });
            }
        }
    }
}
// vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8 :
